/*

Copyright 1992-93 by the University of Edinburgh, Department of Computer Science

Permission to use, copy, modify, distribute, and sell this software and its
documentation for any purpose is hereby granted without fee, provided that
the above copyright notice appear in all copies and that both that
copyright notice and this permission notice appear in supporting
documentation, and that the name of the University of Edinburgh not be used
in advertising or publicity pertaining to distribution of the software
without specific, written prior permission.  The University of Edinburgh
makes no representations about the suitability of this software for any
purpose.  It is provided "as is" without express or implied warranty.

*/
/* Propagate X authority for nsu, yped and the like.  Grab the caller's
 * authority records, presumably under the real uid.  Later, stuff them back
 * out again, presumably under the effective (set) uid.  Display is assumed to
 * be defined by DISPLAY; HOME and XAUTHORITY are assumed to be correct at the
 * time of call. If DISPLAY is undefined do nothing.
 */

#define MAXAUTH 128
#define MAXAUTHNAME 128
#define MAXHOSTNAME 128

#include <X11/X.h>
#include <X11/Xauth.h>
#include <X11/Xos.h>
#include <X11/Xlib.h>

#include <stdio.h>
#ifdef ultrix
char *getenv( char *__name );
#else
#include <stdlib.h>
#endif
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#if defined(SVR4) && !defined(hpux)
#include <unistd.h>
#include <wait.h>
#endif

#ifdef DEBUG
void putIP(a)
char *a;
{	(void) printf("%d.", *a & 255);	a++;
	(void) printf("%d.", *a & 255);	a++;
	(void) printf("%d.", *a & 255);	a++;
	(void) printf("%d ", *a & 255);
}
#endif

static int n;
static Xauth *matches[MAXAUTH];
static char server[MAXHOSTNAME];
static struct hostent *he;

static int testMatch(a)
Xauth *a;
{	char **al;
	char *na, *nb;
	int n;

	if (a->family == FamilyInternet) {
		al = he->h_addr_list;
		while (*al) {
			if (* (long *) *al == * (long *) a->address) {
				return 1;
			}
			al++;
		}
		return 0;
	} else if (a->family == FamilyLocal) {
		n = a->address_length;
		na = server;
		nb = a->address;
		while (n && *na) {
			if (*na != *nb) return 0;
			n--;  na++;  nb++;
		}
		if (n || *na) return 0;
		return 1;
	} else {
		/* Unknown/unsupported family */
		return 0;
	}
}

static void do_grabXauth()
{	Xauth *a;
	char *authFile;
	FILE *file;
	int i;

#ifdef DEBUG
	(void) printf("grabXauth()\n");
#endif

	authFile = XauFileName();
#ifdef DEBUG
	(void) printf("Auth file is %s, locking...\n", authFile);
#endif
	i = XauLockAuth(authFile, 10, 2, 200);
	if (i == LOCK_ERROR) {
		perror("XauLockAuth");
		return;
	} else if (i == LOCK_TIMEOUT) {
		(void) fprintf(stderr, "XauLockAuth timeout: %s\n", authFile);
		return;
	}

#ifdef DEBUG
	(void) printf("Opening...\n");
#endif
	if ((file = fopen(authFile, "r")) == NULL) {
		(void) fprintf(stderr, "Couldn't open %s\n", authFile);
		(void) XauUnlockAuth(authFile);
		return;
	}

	for (;;) {
		a = XauReadAuth(file);
		if (a == NULL) break;
		if (testMatch(a)) matches[n++] = a;
		else XauDisposeAuth(a);
	}
#ifdef DEBUG
	(void) printf("%d matched\n", n);
#endif

	(void) fclose(file);
	(void) XauUnlockAuth(authFile);
}

static void do_stuffXauth()
{	char *authFile;
	char outFile[MAXAUTHNAME];
	FILE *in, *out;
	Xauth *a;
	int i;

#ifdef DEBUG
	(void) printf("stuffXauth() with %d saved\n", n);
#endif

	authFile = XauFileName();
	(void) strcpy(outFile, authFile);
	(void) strcat(outFile, "-n");
#ifdef DEBUG
	(void) printf("Source authFile %s, output to %s\n", authFile, outFile);
#endif

	i = XauLockAuth(authFile, 10, 2, 200);
	if (i == LOCK_ERROR) {
		perror("XauLockAuth");
		return;
	} else if (i == LOCK_TIMEOUT) {
		(void) fprintf(stderr, "XauLockAuth timeout: %s\n", authFile);
		return;
	}

#ifdef DEBUG
	(void) printf("Source locked, opening files\n");
#endif
	in = fopen(authFile, "r");
#ifdef DEBUG
	if (!in) (void) printf("Null FILE* for source\n");
#endif
	if ((out = fopen(outFile, "w")) == NULL) {
		(void) fprintf(stderr, "Couldn't open %s\n", outFile);
		(void) fclose(in);
		(void) XauUnlockAuth(authFile);
		return;
	}

#ifdef DEBUG
	(void) printf("Writing %d matching record(s)\n", n);
#endif
	while (n) {
		a = matches[--n];
		(void) XauWriteAuth(out, a);
		XauDisposeAuth(a);
	}

#ifdef DEBUG
	(void) printf("Copying unmatching records...\n");
#endif
	if (in) {
		for (;;) {
			a = XauReadAuth(in);
			if (a == NULL) break;
			if (!testMatch(a)) (void) XauWriteAuth(out, a);
			XauDisposeAuth(a);
		}
	}

#ifdef DEBUG
	(void) printf("Copied, closing...");
#endif
	(void) fclose(in);
	(void) fclose(out);
	(void) rename(outFile, authFile);
	(void) chmod(authFile, 0600);
	(void) XauUnlockAuth(authFile);
}


#if defined(SVR4) && !defined(hpux)
/* This is a disgusting hack, but there doesn't seem to be any other way to
 * get back to where we were if the euid == 0
 */

static void DoIt(what)
void (*what)();
{	uid_t ruid, euid;
	pid_t child;
	siginfo_t info;

	ruid = getuid();
	euid = geteuid();
#ifdef DEBUG
	(void) fprintf(stderr, "SYSV DoIt, real %d, effective %d\n", ruid, euid);
#endif
	if (ruid != euid) {
		/* We can't just setuid() here, since if our effective uid
		 * were root, which is quite likely, then there would be
		 * no way back (yuk!!).  So... we vfork() and grab the cookies
		 * in the child process.  What a hack!
		 */
		child = vfork();
		if (child < 0) {
			(void) perror("vfork");
			return;
		}

		if (child == 0) {
			/* child, grab the cookies */
			(void) setuid(ruid);
			(*what)();
			_exit(0);
		}

		else {
			/* parent, wait for the child */
			(void) waitid(P_PID, (id_t) child, &info, WEXITED);
		}
	}
	/* else effective uid == real uid, so no need to change */
	else (*what)();
}

#else /* Not SVR4 */

#ifdef hpux
#define setreuid(r, u) setresuid(r, u, -1)
#endif

static void DoIt(what)
void (*what)();
{	uid_t ruid, euid;

	ruid = getuid();
	euid = geteuid();
	(void) fprintf(stderr, "BSD DoIt, real %d, effective %d\n", ruid, euid);

	if (ruid != euid) (void) setreuid(-1, ruid);

	(*what)();

	if (ruid != euid) (void) setreuid(-1, euid);
}

#endif /* not SVR4 */


void grabXauth()
{	char *ch;
	Display *display;

	if (!(ch = getenv("DISPLAY"))) return;
#ifdef notdef
	if (!(display = XOpenDisplay(NULL))) return;
	XCloseDisplay(display);
#endif
	(void) strcpy(server, ch);
	if ((ch = index(server, ':')) != NULL) *ch = '\0';
	if (!(*server)) (void) gethostname(server, MAXHOSTNAME);
	he = gethostbyname(server);

	DoIt(do_grabXauth);
}

void stuffXauth()
{	if (!n) return;  /* Nothing saved, nothing to do */
	DoIt(do_stuffXauth);
}
